/************************************************************************
RWhois Software 

Copyright (c) 1994 Scott Williamson and Mark Kosters

Users may distribute the binary and source code to third parties provided
that the copyright notice and this statement appears on all copies and 
that no charge is made for such copies.  

This software is provided AS IS with no warranties of any kind.  The authors
shall have no liability with respect to the infringement of copyrights,
trade secrets or any patents by this file or any part thereof.  In no
event will the authors be liable for any lost revenue or profits or
other special, indirect and consequential damages.

Comments and additions should be sent to the authors:

                    Scott Williamson
                    Network Solutions
                    505 Huntmar Park Drive
                    Herndon, VA 22071
                    (703) 742-4820
                    scottw@internic.net

                    Mark Kosters
                    Network Solutions
                    505 Huntmar Park Drive
                    Herndon, VA 22071
                    (703) 742-4795
                    markk@internic.net


Developed By: 
  Mark Kosters
***************************************************************************/

#include <stdio.h>
#include <string.h>
#include "defines.h"
#include "conf.h"
#include "attributes.h"
#include "records.h"
#include "parse.h"
#include "objects.h"
#include "whois.h"
#include "client_msgs.h"

/***
  external stuff
****/
#ifdef USE_WAIS
record_struct *wais_interface();
#endif
#ifdef USE_MKDB
record_struct *mkdb_interface();
#endif

field_struct *read_fields_from_text();

/***
  local prototypes
***/
static int can_do_more_searches ();
static record_struct *run_search();
static record_struct *get_root_server_records();

/******************************************************************************
  this is the main ball of wax
  returns 0 if standard query with holdconnect = FALSE
          1 if an enviroment command or holdconnect = TRUE
if internal_see_also set, does not reduce and returns TRUE if found, false if
   not (for internal see-also's).

pass 0 is for regular and cached data
pass 1 is for referral (no reduction)
pass 2+ is for referrals with reductions
******************************************************************************/
whois(query_str, root_server, internal_see_also)
  char *query_str;
  int root_server;
  int internal_see_also;

{
  record_struct *records;
  llist_struct *list_of_objects;
  query_item_struct *query_list;
  int reduce = FALSE, pass = 0, hits = FALSE, next_action;
  object_type search_object_type;
  char *display_mode;

  trim(query_str);
  log_entry (RWHOIS_LOG_FILE, "Pass 0 Arg: %s", query_str);
  save_original_query (query_str);

  /***
    okay this is for those who type return or other bad things
    ***/

  if (!strlen (query_str) || (query_list = encode_command(query_str)) == NULL) {
    if (holdconnect_status ()) {
      help();
      return (TRUE);
    }
    else { /* if no holdconnect exit as if regular query */
      help();
      return (FALSE);
    }
  }

  /***
    this is for the directives only
    ***/

  if (is_directive (query_list)) {
    run_directive (query_list);
    return (holdconnect_status());
  }

  /****
    does the search/display for search commands
    ****/

  while (can_do_more_searches (hits, pass, &search_object_type, &reduce, 
                               &next_action, query_list, internal_see_also)) {
    if (reduce)
      reduce_query(query_list, pass);
    hits = run_query (search_object_type, FALSE, pass, query_list,
                      internal_see_also);
    record_results (FALSE, reduce, hits);
    if (cache_status()) {
      int old_hits = hits;
      hits += run_query (ALL_OBJECTS, TRUE, pass, query_list, 
                         internal_see_also);
      record_results (TRUE, FALSE, hits-old_hits);
    }
    pass++;
  } 
  /**
    Done with the searching - now do what next_action says to do
    **/

  switch (next_action) {
  case PUNT_ACTION:
    if (!root_server) {
      if ((records = get_root_server_records()) != NULL)
        log_entry (RWHOIS_LOG_FILE, "ROOT QUERY");
      display_results (records, FALSE, "summary", NULL, FALSE);
      print_ok();
    }
    else if (!hits)
      print_error (NO_RECORDS_FOUND, "");
    break;

  case WHOIS_PP_ACTION:
    printf ("%%info on\n");
    printf ("This query is not in our database and appears to be non-reducible\n");
    printf ("-- OR --\n");
    printf ("we do not have a referral for this information.\n");
    printf ("%%info off\n");
    print_error (NO_RECORDS_FOUND, "");
    break;

  case DONE_ACTION:
    if (!internal_see_also)
      print_ok ();
    break;
  }
  if (!internal_see_also) 
    return (holdconnect_status());
  else 
    return (hits);
}

/******************************************************************************
  this procedure sees if ok to keep on calling the search loop
  returns TRUE to keep on going
          FALSE otherwise
******************************************************************************/
static int can_do_more_searches (hits, pass, search_object_type, reduce, 
                                 action, query_list, internal_see_also)
  int hits;
  int pass;
  object_type *search_object_type;
  int *reduce;
  int *action;
  query_item_struct *query_list;
  int internal_see_also;

{
  /***
    for now, if any hits - exit out
    ***/
  if (hits > 0) {
    *action = DONE_ACTION;
    return (FALSE);
  }
  /***
    for now, leave this be
    ***/
  if (pass > 0) {
    if (internal_see_also) {
      *action = DONE_ACTION;
      return (FALSE);
    }
    *reduce =  TRUE;
    *search_object_type = REFERRAL;
    if ((*action = is_query_reducible (query_list, pass)) != KEEPON_ACTION)
      return (FALSE);
  }
  else
    *search_object_type = ALL_OBJECTS;
  return (TRUE);
}

/******************************************************************************
  this procedure does query - calling the search and display modules
  returns number of hits
******************************************************************************/
int run_query(search_object_type, cache, pass, query_list, sub_query)
  object_type search_object_type;
  int cache;
  int pass;
  query_item_struct *query_list;
  int sub_query;

{
  record_struct *records;
  int hits;

  records = run_search(search_object_type, cache, pass, query_list);
  hits = display_results(records, cache, get_display_mode(query_list), 
                         decode_data_for_authority_element (query_list),
                         sub_query);
  if (hits)
    free_records(records);
  return (hits);
}

/******************************************************************************
  this procedure runs the search part of the query
******************************************************************************/
static record_struct *run_search(search_object_type, cache, pass, 
                                 query_list)
  object_type search_object_type;
  int cache;
  int pass;
  query_item_struct *query_list;

{
  llist_struct *list_of_objects;
  char query_str[MAX_LINE];
  object_struct *object;
  record_struct *records = NULL;
  char *db_fname;
  int result;
  int mkdb_error_code = 0;

  list_of_objects = setup_search_objects(query_list, search_object_type);
  decode_command_list(query_str, query_list);
  if (pass > 0)
    log_entry (RWHOIS_LOG_FILE, "Pass %d Arg: %s", pass, query_str);

  while (list_of_objects != NULL) {
    object = list_of_objects->datum;
    if (cache && object->cache_db != NULL)
      db_fname = object->cache_db;
    else
      db_fname = object->local_db;
    if (debug_status()) {
      printf ("run_search: trying template |%s| file |%s|\n", object->name,
              db_fname);
      printf ("run_search: looking for |%s|\n", query_str);
    }
#ifdef USE_WAIS
    printf ("in wais\n");
    records = wais_interface ((char *)NULL, 0, records,  &result, query_str, 
                              db_fname, get_max_hit_limit());
#endif
#ifdef USE_MKDB
    records = mkdb_interface(db_fname, query_str, get_max_hit_limit(), 
                             (search_object_type == REFERRAL? MAX_REFERRAL_LINES : 0),
                             &mkdb_error_code, records);
#endif
    list_of_objects = list_of_objects->nxt;
  }
  return (records);
}

/******************************************************************************
  this procedure runs the search part of the query
******************************************************************************/
void record_results (cache, reduce, hits)
  int cache;
  int reduce;
  int hits;

{
  if (cache)
    log_entry (RWHOIS_LOG_FILE, "Ans: %d cache hits", hits);
  else if (reduce)
    log_entry (RWHOIS_LOG_FILE, "Ans: %d referral hits", hits);
  else
    log_entry (RWHOIS_LOG_FILE, "Ans: %d data hits", hits);
}

/******************************************************************************
  this procedure returns the root server record
  now it is hard-coded - this will change
  bug fixes thanks to drc@nis.apnic.net 4/29/94
******************************************************************************/
static record_struct *get_root_server_records()

{
  char txt[MAX_LINE*5];
  char line[MAX_LINE];
  record_struct *top = NULL, *new, *cur;
  FILE *fp;

  *txt = '\0';

  if ((fp = fopen (DEFAULT_RWHOIS_ROOT_FILE, "r")) == NULL)
    print_error (MISC_DIAG, "Can not open root server file '%s'", 
                 DEFAULT_RWHOIS_ROOT_FILE);

  while ( fgets (line, MAX_LINE, fp) != NULL) {
    if (strncmp (line,"-----", 5) == 0) {
      if (*txt != '\0') {
        new = NEW_RECORD;
        new->fields = read_fields_from_text (txt);
        if (top == NULL)
          cur = top = new;
        else {
          cur->nxt = new;
          cur = new;
        }
        *txt = '\0';
      }
    }
    else
      strncat (txt, line, MAX_LINE*5);
  }
  fclose (fp);
  if (txt != NULL)
  {
    new = NEW_RECORD;
    new->fields = read_fields_from_text (txt);
    if (top == NULL)
      cur = top = new;
    else {
      cur->nxt = new;
      cur = new;
    }
  }
  return (top);
}




